@@ -24,14 +24,11 @@
 #include <libnftnl/expr.h>
 #include <libnftnl/rule.h>
 
-#include "expr_ops.h"
-#include <buffer.h>
-
 #ifndef IPPROTO_MH
 #define IPPROTO_MH 135
 #endif
 
-struct nft_expr_exthdr {
+struct nftnl_expr_exthdr {
 	enum nft_registers	dreg;
 	uint32_t		offset;
 	uint32_t		len;
@@ -39,22 +36,22 @@ struct nft_expr_exthdr {
 };
 
 static int
-nft_rule_expr_exthdr_set(struct nft_rule_expr *e, uint16_t type,
+nftnl_expr_exthdr_set(struct nftnl_expr *e, uint16_t type,
 			  const void *data, uint32_t data_len)
 {
-	struct nft_expr_exthdr *exthdr = nft_expr_data(e);
+	struct nftnl_expr_exthdr *exthdr = nftnl_expr_data(e);
 
 	switch(type) {
-	case NFT_EXPR_EXTHDR_DREG:
+	case NFTNL_EXPR_EXTHDR_DREG:
 		exthdr->dreg = *((uint32_t *)data);
 		break;
-	case NFT_EXPR_EXTHDR_TYPE:
+	case NFTNL_EXPR_EXTHDR_TYPE:
 		exthdr->type = *((uint8_t *)data);
 		break;
-	case NFT_EXPR_EXTHDR_OFFSET:
+	case NFTNL_EXPR_EXTHDR_OFFSET:
 		exthdr->offset = *((uint32_t *)data);
 		break;
-	case NFT_EXPR_EXTHDR_LEN:
+	case NFTNL_EXPR_EXTHDR_LEN:
 		exthdr->len = *((uint32_t *)data);
 		break;
 	default:
@@ -64,29 +61,29 @@ nft_rule_expr_exthdr_set(struct nft_rule_expr *e, uint16_t type,
 }
 
 static const void *
-nft_rule_expr_exthdr_get(const struct nft_rule_expr *e, uint16_t type,
+nftnl_expr_exthdr_get(const struct nftnl_expr *e, uint16_t type,
 			 uint32_t *data_len)
 {
-	struct nft_expr_exthdr *exthdr = nft_expr_data(e);
+	struct nftnl_expr_exthdr *exthdr = nftnl_expr_data(e);
 
 	switch(type) {
-	case NFT_EXPR_EXTHDR_DREG:
+	case NFTNL_EXPR_EXTHDR_DREG:
 		*data_len = sizeof(exthdr->dreg);
 		return &exthdr->dreg;
-	case NFT_EXPR_EXTHDR_TYPE:
+	case NFTNL_EXPR_EXTHDR_TYPE:
 		*data_len = sizeof(exthdr->type);
 		return &exthdr->type;
-	case NFT_EXPR_EXTHDR_OFFSET:
+	case NFTNL_EXPR_EXTHDR_OFFSET:
 		*data_len = sizeof(exthdr->offset);
 		return &exthdr->offset;
-	case NFT_EXPR_EXTHDR_LEN:
+	case NFTNL_EXPR_EXTHDR_LEN:
 		*data_len = sizeof(exthdr->len);
 		return &exthdr->len;
 	}
 	return NULL;
 }
 
-static int nft_rule_expr_exthdr_cb(const struct nlattr *attr, void *data)
+static int nftnl_expr_exthdr_cb(const struct nlattr *attr, void *data)
 {
 	const struct nlattr **tb = data;
 	int type = mnl_attr_get_type(attr);
@@ -112,44 +109,44 @@ static int nft_rule_expr_exthdr_cb(const struct nlattr *attr, void *data)
 }
 
 static void
-nft_rule_expr_exthdr_build(struct nlmsghdr *nlh, struct nft_rule_expr *e)
+nftnl_expr_exthdr_build(struct nlmsghdr *nlh, const struct nftnl_expr *e)
 {
-	struct nft_expr_exthdr *exthdr = nft_expr_data(e);
+	struct nftnl_expr_exthdr *exthdr = nftnl_expr_data(e);
 
-	if (e->flags & (1 << NFT_EXPR_EXTHDR_DREG))
+	if (e->flags & (1 << NFTNL_EXPR_EXTHDR_DREG))
 		mnl_attr_put_u32(nlh, NFTA_EXTHDR_DREG, htonl(exthdr->dreg));
-	if (e->flags & (1 << NFT_EXPR_EXTHDR_TYPE))
+	if (e->flags & (1 << NFTNL_EXPR_EXTHDR_TYPE))
 		mnl_attr_put_u8(nlh, NFTA_EXTHDR_TYPE, exthdr->type);
-	if (e->flags & (1 << NFT_EXPR_EXTHDR_OFFSET))
+	if (e->flags & (1 << NFTNL_EXPR_EXTHDR_OFFSET))
 		mnl_attr_put_u32(nlh, NFTA_EXTHDR_OFFSET, htonl(exthdr->offset));
-	if (e->flags & (1 << NFT_EXPR_EXTHDR_LEN))
+	if (e->flags & (1 << NFTNL_EXPR_EXTHDR_LEN))
 		mnl_attr_put_u32(nlh, NFTA_EXTHDR_LEN, htonl(exthdr->len));
 }
 
 static int
-nft_rule_expr_exthdr_parse(struct nft_rule_expr *e, struct nlattr *attr)
+nftnl_expr_exthdr_parse(struct nftnl_expr *e, struct nlattr *attr)
 {
-	struct nft_expr_exthdr *exthdr = nft_expr_data(e);
+	struct nftnl_expr_exthdr *exthdr = nftnl_expr_data(e);
 	struct nlattr *tb[NFTA_EXTHDR_MAX+1] = {};
 
-	if (mnl_attr_parse_nested(attr, nft_rule_expr_exthdr_cb, tb) < 0)
+	if (mnl_attr_parse_nested(attr, nftnl_expr_exthdr_cb, tb) < 0)
 		return -1;
 
 	if (tb[NFTA_EXTHDR_DREG]) {
 		exthdr->dreg = ntohl(mnl_attr_get_u32(tb[NFTA_EXTHDR_DREG]));
-		e->flags |= (1 << NFT_EXPR_EXTHDR_DREG);
+		e->flags |= (1 << NFTNL_EXPR_EXTHDR_DREG);
 	}
 	if (tb[NFTA_EXTHDR_TYPE]) {
 		exthdr->type = mnl_attr_get_u8(tb[NFTA_EXTHDR_TYPE]);
-		e->flags |= (1 << NFT_EXPR_EXTHDR_TYPE);
+		e->flags |= (1 << NFTNL_EXPR_EXTHDR_TYPE);
 	}
 	if (tb[NFTA_EXTHDR_OFFSET]) {
 		exthdr->offset = ntohl(mnl_attr_get_u32(tb[NFTA_EXTHDR_OFFSET]));
-		e->flags |= (1 << NFT_EXPR_EXTHDR_OFFSET);
+		e->flags |= (1 << NFTNL_EXPR_EXTHDR_OFFSET);
 	}
 	if (tb[NFTA_EXTHDR_LEN]) {
 		exthdr->len = ntohl(mnl_attr_get_u32(tb[NFTA_EXTHDR_LEN]));
-		e->flags |= (1 << NFT_EXPR_EXTHDR_LEN);
+		e->flags |= (1 << NFTNL_EXPR_EXTHDR_LEN);
 	}
 
 	return 0;
@@ -190,32 +187,32 @@ static inline int str2exthdr_type(const char *str)
 }
 
 static int
-nft_rule_expr_exthdr_json_parse(struct nft_rule_expr *e, json_t *root,
-				struct nft_parse_err *err)
+nftnl_expr_exthdr_json_parse(struct nftnl_expr *e, json_t *root,
+				struct nftnl_parse_err *err)
 {
 #ifdef JSON_PARSING
 	const char *exthdr_type;
 	uint32_t uval32;
 	int type;
 
-	if (nft_jansson_parse_reg(root, "dreg", NFT_TYPE_U32, &uval32,
+	if (nftnl_jansson_parse_reg(root, "dreg", NFTNL_TYPE_U32, &uval32,
 				  err) == 0)
-		nft_rule_expr_set_u32(e, NFT_EXPR_EXTHDR_DREG, uval32);
+		nftnl_expr_set_u32(e, NFTNL_EXPR_EXTHDR_DREG, uval32);
 
-	exthdr_type = nft_jansson_parse_str(root, "exthdr_type", err);
+	exthdr_type = nftnl_jansson_parse_str(root, "exthdr_type", err);
 	if (exthdr_type != NULL) {
 		type = str2exthdr_type(exthdr_type);
 		if (type < 0)
 			return -1;
-		nft_rule_expr_set_u32(e, NFT_EXPR_EXTHDR_TYPE, type);
+		nftnl_expr_set_u32(e, NFTNL_EXPR_EXTHDR_TYPE, type);
 	}
 
-	if (nft_jansson_parse_val(root, "offset", NFT_TYPE_U32, &uval32,
+	if (nftnl_jansson_parse_val(root, "offset", NFTNL_TYPE_U32, &uval32,
 				  err) == 0)
-		nft_rule_expr_set_u32(e, NFT_EXPR_EXTHDR_OFFSET, uval32);
+		nftnl_expr_set_u32(e, NFTNL_EXPR_EXTHDR_OFFSET, uval32);
 
-	if (nft_jansson_parse_val(root, "len", NFT_TYPE_U32, &uval32, err) == 0)
-		nft_rule_expr_set_u32(e, NFT_EXPR_EXTHDR_LEN, uval32);
+	if (nftnl_jansson_parse_val(root, "len", NFTNL_TYPE_U32, &uval32, err) == 0)
+		nftnl_expr_set_u32(e, NFTNL_EXPR_EXTHDR_LEN, uval32);
 
 	return 0;
 #else
@@ -224,67 +221,28 @@ nft_rule_expr_exthdr_json_parse(struct nft_rule_expr *e, json_t *root,
 #endif
 }
 
-static int
-nft_rule_expr_exthdr_xml_parse(struct nft_rule_expr *e, mxml_node_t *tree,
-			       struct nft_parse_err *err)
+static int nftnl_expr_exthdr_export(char *buf, size_t len,
+				    const struct nftnl_expr *e, int type)
 {
-#ifdef XML_PARSING
-	const char *exthdr_type;
-	int type;
-	uint32_t dreg, len, offset;
+	struct nftnl_expr_exthdr *exthdr = nftnl_expr_data(e);
+	NFTNL_BUF_INIT(b, buf, len);
 
-	if (nft_mxml_reg_parse(tree, "dreg", &dreg, MXML_DESCEND_FIRST,
-			       NFT_XML_MAND, err) == 0)
-		nft_rule_expr_set_u32(e, NFT_EXPR_EXTHDR_DREG, dreg);
+	if (e->flags & (1 << NFTNL_EXPR_EXTHDR_DREG))
+		nftnl_buf_u32(&b, type, exthdr->dreg, DREG);
+	if (e->flags & (1 << NFTNL_EXPR_EXTHDR_TYPE))
+		nftnl_buf_str(&b, type, type2str(exthdr->type), EXTHDR_TYPE);
+	if (e->flags & (1 << NFTNL_EXPR_EXTHDR_OFFSET))
+		nftnl_buf_u32(&b, type, exthdr->offset, OFFSET);
+	if (e->flags & (1 << NFTNL_EXPR_EXTHDR_LEN))
+		nftnl_buf_u32(&b, type, exthdr->len, LEN);
 
-	exthdr_type = nft_mxml_str_parse(tree, "exthdr_type",
-					 MXML_DESCEND_FIRST, NFT_XML_MAND, err);
-	if (exthdr_type != NULL) {
-		type = str2exthdr_type(exthdr_type);
-		if (type < 0)
-			return -1;
-		nft_rule_expr_set_u8(e, NFT_EXPR_EXTHDR_TYPE, type);
-	}
-
-	/* Get and set <offset> */
-	if (nft_mxml_num_parse(tree, "offset", MXML_DESCEND_FIRST, BASE_DEC,
-			       &offset, NFT_TYPE_U32, NFT_XML_MAND, err) == 0)
-		nft_rule_expr_set_u32(e, NFT_EXPR_EXTHDR_OFFSET, offset);
-
-	/* Get and set <len> */
-	if (nft_mxml_num_parse(tree, "len", MXML_DESCEND_FIRST, BASE_DEC,
-			       &len, NFT_TYPE_U32, NFT_XML_MAND, err) == 0)
-		nft_rule_expr_set_u32(e, NFT_EXPR_EXTHDR_LEN, len);
-
-	return 0;
-#else
-	errno = EOPNOTSUPP;
-	return -1;
-#endif
+	return nftnl_buf_done(&b);
 }
 
-static int nft_rule_expr_exthdr_export(char *buf, size_t len,
-				       struct nft_rule_expr *e, int type)
+static int nftnl_expr_exthdr_snprintf_default(char *buf, size_t len,
+					      const struct nftnl_expr *e)
 {
-	struct nft_expr_exthdr *exthdr = nft_expr_data(e);
-	NFT_BUF_INIT(b, buf, len);
-
-	if (e->flags & (1 << NFT_EXPR_EXTHDR_DREG))
-		nft_buf_u32(&b, type, exthdr->dreg, DREG);
-	if (e->flags & (1 << NFT_EXPR_EXTHDR_TYPE))
-		nft_buf_str(&b, type, type2str(exthdr->type), EXTHDR_TYPE);
-	if (e->flags & (1 << NFT_EXPR_EXTHDR_OFFSET))
-		nft_buf_u32(&b, type, exthdr->offset, OFFSET);
-	if (e->flags & (1 << NFT_EXPR_EXTHDR_LEN))
-		nft_buf_u32(&b, type, exthdr->len, LEN);
-
-	return nft_buf_done(&b);
-}
-
-static int nft_rule_expr_exthdr_snprintf_default(char *buf, size_t len,
-						 struct nft_rule_expr *e)
-{
-	struct nft_expr_exthdr *exthdr = nft_expr_data(e);
+	struct nftnl_expr_exthdr *exthdr = nftnl_expr_data(e);
 
 	return snprintf(buf, len, "load %ub @ %u + %u => reg %u ",
 			exthdr->len, exthdr->type, exthdr->offset,
@@ -292,35 +250,49 @@ static int nft_rule_expr_exthdr_snprintf_default(char *buf, size_t len,
 }
 
 static int
-nft_rule_expr_exthdr_snprintf(char *buf, size_t len, uint32_t type,
-			       uint32_t flags, struct nft_rule_expr *e)
+nftnl_expr_exthdr_snprintf(char *buf, size_t len, uint32_t type,
+			   uint32_t flags, const struct nftnl_expr *e)
 {
 	switch (type) {
-	case NFT_OUTPUT_DEFAULT:
-		return nft_rule_expr_exthdr_snprintf_default(buf, len, e);
-	case NFT_OUTPUT_XML:
-	case NFT_OUTPUT_JSON:
-		return nft_rule_expr_exthdr_export(buf, len, e, type);
+	case NFTNL_OUTPUT_DEFAULT:
+		return nftnl_expr_exthdr_snprintf_default(buf, len, e);
+	case NFTNL_OUTPUT_XML:
+	case NFTNL_OUTPUT_JSON:
+		return nftnl_expr_exthdr_export(buf, len, e, type);
 	default:
 		break;
 	}
 	return -1;
 }
 
+static bool nftnl_expr_exthdr_cmp(const struct nftnl_expr *e1,
+				  const struct nftnl_expr *e2)
+{
+	struct nftnl_expr_exthdr *h1 = nftnl_expr_data(e1);
+	struct nftnl_expr_exthdr *h2 = nftnl_expr_data(e2);
+	bool eq = true;
+
+	if (e1->flags & (1 << NFTNL_EXPR_EXTHDR_DREG))
+		eq &= (h1->dreg == h2->dreg);
+	if (e1->flags & (1 << NFTNL_EXPR_EXTHDR_OFFSET))
+		eq &= (h1->offset == h2->offset);
+	if (e1->flags & (1 << NFTNL_EXPR_EXTHDR_LEN))
+		eq &= (h1->len == h2->len);
+	if (e1->flags & (1 << NFTNL_EXPR_EXTHDR_TYPE))
+		eq &= (h1->type == h2->type);
+
+	return eq;
+}
+
 struct expr_ops expr_ops_exthdr = {
 	.name		= "exthdr",
-	.alloc_len	= sizeof(struct nft_expr_exthdr),
+	.alloc_len	= sizeof(struct nftnl_expr_exthdr),
 	.max_attr	= NFTA_EXTHDR_MAX,
-	.set		= nft_rule_expr_exthdr_set,
-	.get		= nft_rule_expr_exthdr_get,
-	.parse		= nft_rule_expr_exthdr_parse,
-	.build		= nft_rule_expr_exthdr_build,
-	.snprintf	= nft_rule_expr_exthdr_snprintf,
-	.xml_parse	= nft_rule_expr_exthdr_xml_parse,
-	.json_parse	= nft_rule_expr_exthdr_json_parse,
+	.cmp		= nftnl_expr_exthdr_cmp,
+	.set		= nftnl_expr_exthdr_set,
+	.get		= nftnl_expr_exthdr_get,
+	.parse		= nftnl_expr_exthdr_parse,
+	.build		= nftnl_expr_exthdr_build,
+	.snprintf	= nftnl_expr_exthdr_snprintf,
+	.json_parse	= nftnl_expr_exthdr_json_parse,
 };
-
-static void __init expr_exthdr_init(void)
-{
-	nft_expr_ops_register(&expr_ops_exthdr);
-}
